@@ -25,9 +25,14 @@ import sys
 import re
 import functools
 from collections import namedtuple
+import bracex
 from . import _wcparse
 from . import _wcmatch
 from . import util
+from typing import (
+    Optional, Iterator, Iterable, List, AnyStr, Union, Generic,
+    Tuple, Pattern, Callable, Any, Set, Sequence, cast
+)
 
 __all__ = (
     "CASE", "IGNORECASE", "RAWCHARS", "DOTGLOB", "DOTMATCH",
@@ -102,19 +107,18 @@ FLAG_MASK = (
     _NOABSOLUTE
 )
 
-_RE_PATHLIB_DOT_NORM = [
+_RE_PATHLIB_DOT_NORM = (
     re.compile(r'(?:((?<=^)|(?<=/))\.(?:/|$))+'),
     re.compile(br'(?:((?<=^)|(?<=/))\.(?:/|$))+')
+)  # type: Tuple[Pattern[str], Pattern[bytes]]
 
-]
-
-_RE_WIN_PATHLIB_DOT_NORM = [
+_RE_WIN_PATHLIB_DOT_NORM = (
     re.compile(r'(?:((?<=^)|(?<=[\\/]))\.(?:[\\/]|$))+'),
     re.compile(br'(?:((?<=^)|(?<=[\\/]))\.(?:[\\/]|$))+')
-]
+)  # type: Tuple[Pattern[str], Pattern[bytes]]
 
 
-def _flag_transform(flags):
+def _flag_transform(flags: int) -> int:
     """Transform flags to glob defaults."""
 
     # Enabling both cancels out
@@ -135,11 +139,13 @@ def _flag_transform(flags):
     return flags
 
 
-class _GlobPart(namedtuple('_GlobPart', ['pattern', 'is_magic', 'is_globstar', 'dir_only', 'is_drive'])):
+class _GlobPart(
+    namedtuple('_GlobPart', ['pattern', 'is_magic', 'is_globstar', 'dir_only', 'is_drive']),
+):
     """File Glob."""
 
 
-class _GlobSplit(object):
+class _GlobSplit(Generic[AnyStr]):
     """
     Split glob pattern on "magic" file and directories.
 
@@ -166,12 +172,12 @@ class _GlobSplit(object):
 
     """
 
-    def __init__(self, pattern, flags):
+    def __init__(self, pattern: AnyStr, flags: int) -> None:
         """Initialize."""
 
+        self.pattern = pattern  # type: AnyStr
         self.unix = _wcparse.is_unix_style(flags)
         self.flags = flags
-        self.pattern = pattern
         self.no_abs = bool(flags & _wcparse._NOABSOLUTE)
         self.globstar = bool(flags & GLOBSTAR)
         self.matchbase = bool(flags & MATCHBASE)
@@ -185,7 +191,6 @@ class _GlobSplit(object):
         if flags & NEGATE:
             flags ^= NEGATE
         self.flags = flags
-        self.is_bytes = isinstance(pattern, bytes)
         self.extend = bool(flags & EXTMATCH)
         if not self.unix:
             self.win_drive_detect = True
@@ -197,10 +202,9 @@ class _GlobSplit(object):
             self.sep = '/'
         # Once split, Windows file names will never have `\\` in them,
         # so we can use the Unix magic detect
-        ptype = util.BYTES if self.is_bytes else util.UNICODE
-        self.magic_symbols = _wcparse._get_magic_symbols(ptype, self.unix, self.flags)[0]
+        self.magic_symbols = _wcparse._get_magic_symbols(pattern, self.unix, self.flags)[0]  # type: Set[AnyStr]
 
-    def is_magic(self, name):
+    def is_magic(self, name: AnyStr) -> bool:
         """Check if name contains magic characters."""
 
         for c in self.magic_symbols:
@@ -208,7 +212,7 @@ class _GlobSplit(object):
                 return True
         return False
 
-    def _sequence(self, i):
+    def _sequence(self, i: util.StringIter) -> None:
         """Handle character group."""
 
         c = next(i)
@@ -228,7 +232,7 @@ class _GlobSplit(object):
                 raise StopIteration
             c = next(i)
 
-    def _references(self, i, sequence=False):
+    def _references(self, i: util.StringIter, sequence: bool = False) -> str:
         """Handle references."""
 
         value = ''
@@ -249,7 +253,7 @@ class _GlobSplit(object):
             pass
         return value
 
-    def parse_extend(self, c, i):
+    def parse_extend(self, c: str, i: util.StringIter) -> bool:
         """Parse extended pattern lists."""
 
         # Start list parsing
@@ -285,7 +289,7 @@ class _GlobSplit(object):
 
         return success
 
-    def store(self, value, l, dir_only):
+    def store(self, value: AnyStr, l: List[_GlobPart], dir_only: bool) -> None:
         """Group patterns by literals and potential magic patterns."""
 
         if l and value in (b'', ''):
@@ -294,39 +298,43 @@ class _GlobSplit(object):
         globstar = value in (b'**', '**') and self.globstar
         magic = self.is_magic(value)
         if magic:
-            value = _wcparse._compile(value, self.flags)
-        if globstar and l and l[-1].is_globstar:
-            l[-1] = _GlobPart(value, magic, globstar, dir_only, False)
+            v = cast(Pattern[AnyStr], _wcparse._compile(value, self.flags))  # type: Union[Pattern[AnyStr], AnyStr]
         else:
-            l.append(_GlobPart(value, magic, globstar, dir_only, False))
+            v = value
+        if globstar and l and l[-1].is_globstar:
+            l[-1] = _GlobPart(v, magic, globstar, dir_only, False)
+        else:
+            l.append(_GlobPart(v, magic, globstar, dir_only, False))
 
-    def split(self):
+    def split(self) -> List[_GlobPart]:
         """Start parsing the pattern."""
 
         split_index = []
         parts = []
         start = -1
 
-        pattern = self.pattern.decode('latin-1') if self.is_bytes else self.pattern
+        if isinstance(self.pattern, bytes):
+            is_bytes = True
+            pattern = self.pattern.decode('latin-1')
+        else:
+            is_bytes = False
+            pattern = self.pattern
 
         i = util.StringIter(pattern)
-        iter(i)
 
         # Detect and store away windows drive as a literal
         if self.win_drive_detect:
             root_specified, drive, slash, end = _wcparse._get_win_drive(pattern)
             if drive is not None:
-                if self.is_bytes:
-                    drive = drive.encode('latin-1')
-                parts.append(_GlobPart(drive, False, False, True, True))
+                parts.append(_GlobPart(drive.encode('latin-1') if is_bytes else drive, False, False, True, True))
                 start = end - 1
                 i.advance(start)
             elif drive is None and root_specified:
-                parts.append(_GlobPart(b'\\' if self.is_bytes else '\\', False, False, True, True))
+                parts.append(_GlobPart(b'\\' if is_bytes else '\\', False, False, True, True))
                 start = 1
                 i.advance(2)
         elif not self.win_drive_detect and pattern.startswith('/'):
-            parts.append(_GlobPart(b'/' if self.is_bytes else '/', False, False, True, True))
+            parts.append(_GlobPart(b'/' if is_bytes else '/', False, False, True, True))
             start = 0
             i.advance(1)
 
@@ -353,30 +361,24 @@ class _GlobSplit(object):
                     i.rewind(i.index - index)
 
         for split, offset in split_index:
-            if self.is_bytes:
-                value = pattern[start + 1:split].encode('latin-1')
-            else:
-                value = pattern[start + 1:split]
-            self.store(value, parts, True)
+            value = pattern[start + 1:split]
+            self.store(cast(AnyStr, value.encode('latin-1') if is_bytes else value), parts, True)
             start = split + offset
 
         if start < len(pattern):
-            if self.is_bytes:
-                value = pattern[start + 1:].encode('latin-1')
-            else:
-                value = pattern[start + 1:]
+            value = pattern[start + 1:]
             if value:
-                self.store(value, parts, False)
+                self.store(cast(AnyStr, value.encode('latin-1') if is_bytes else value), parts, False)
 
         if len(pattern) == 0:
-            parts.append(_GlobPart(pattern.encode('latin-1') if self.is_bytes else pattern, False, False, False, False))
+            parts.append(_GlobPart(pattern.encode('latin-1') if is_bytes else pattern, False, False, False, False))
 
         if (
             (self.extmatchbase and not parts[0].is_drive) or
             (self.matchbase and len(parts) == 1 and not parts[0].dir_only)
         ):
             self.globstar = True
-            parts.insert(0, _GlobPart(b'**' if self.is_bytes else '**', True, True, True, False))
+            parts.insert(0, _GlobPart(b'**' if is_bytes else '**', True, True, True, False))
 
         if self.no_abs and parts and parts[0].is_drive:
             raise ValueError('The pattern must be a relative path pattern')
@@ -384,75 +386,92 @@ class _GlobSplit(object):
         return parts
 
 
-class Glob(object):
+class Glob(Generic[AnyStr]):
     """Glob patterns."""
 
-    def __init__(self, pattern, flags=0, root_dir=None, dir_fd=None, limit=_wcparse.PATTERN_LIMIT):
+    def __init__(
+        self,
+        pattern: Union[str, bytes, Sequence[AnyStr]],
+        flags: int = 0,
+        root_dir: Optional[Union[AnyStr, 'os.PathLike[AnyStr]']] = None,
+        dir_fd: Optional[int] = None,
+        limit: int = _wcparse.PATTERN_LIMIT
+    ) -> None:
         """Initialize the directory walker object."""
 
-        self.seen = set()
-        self.is_bytes = isinstance(pattern[0], bytes)
-        self.current = b'.' if self.is_bytes else '.'
-        self.dir_fd = dir_fd if SUPPORT_DIR_FD else None
-        self.root_dir = os.fspath(root_dir) if root_dir is not None else self.current
-        self.nounique = bool(flags & NOUNIQUE)
-        self.mark = bool(flags & MARK)
+        pats = _wcparse.to_str_sequence(pattern)
+
+        self.pattern = []  # type: List[List[_GlobPart]]
+        self.npatterns = []  # type: List[Pattern[AnyStr]]
+        self.seen = set()  # type: Set[AnyStr]
+        self.dir_fd = dir_fd if SUPPORT_DIR_FD else None  # type: Optional[int]
+        self.nounique = bool(flags & NOUNIQUE)  # type: bool
+        self.mark = bool(flags & MARK)  # type: bool
         # Only scan for `.` and `..` if it is specifically requested.
-        self.scandotdir = flags & SCANDOTDIR
+        self.scandotdir = bool(flags & SCANDOTDIR)  # type: bool
         if self.mark:
             flags ^= MARK
-        self.negateall = bool(flags & NEGATEALL)
+        self.negateall = bool(flags & NEGATEALL)  # type: bool
         if self.negateall:
             flags ^= NEGATEALL
-        self.nodir = bool(flags & NODIR)
+        self.nodir = bool(flags & NODIR)  # type: bool
         if self.nodir:
             flags ^= NODIR
-        self.pathlib = bool(flags & _PATHLIB)
+        self.pathlib = bool(flags & _PATHLIB)  # type: bool
         if self.pathlib:
             flags ^= _PATHLIB
         # Right to left searching is only for matching
         if flags & _RTL:  # pragma: no cover
             flags ^= _RTL
-        self.flags = _flag_transform(flags | REALPATH)
-        self.negate_flags = self.flags
+        self.flags = _flag_transform(flags | REALPATH)  # type: int
+        self.negate_flags = self.flags  # type: int
         if not self.scandotdir and not self.flags & NODOTDIR:
             self.flags |= NODOTDIR
-        self.raw_chars = bool(self.flags & RAWCHARS)
-        self.follow_links = bool(self.flags & FOLLOW)
-        self.dot = bool(self.flags & DOTMATCH)
-        self.unix = not bool(self.flags & FORCEWIN)
-        self.negate = bool(self.flags & NEGATE)
-        self.globstar = bool(self.flags & GLOBSTAR)
-        self.braces = bool(self.flags & BRACE)
-        self.matchbase = bool(self.flags & MATCHBASE)
-        self.case_sensitive = _wcparse.get_case(self.flags)
-        self.specials = (b'.', b'..') if self.is_bytes else ('.', '..')
-        self.empty = b'' if self.is_bytes else ''
-        self.stars = b'**' if self.is_bytes else '**'
-        self.limit = limit
-        if self.flags & FORCEWIN:
-            self.sep = b'\\' if self.is_bytes else '\\'
-            self.seps = (b'/' if self.is_bytes else '/', self.sep)
-            self.re_pathlib_norm = _RE_WIN_PATHLIB_DOT_NORM[util.BYTES if self.is_bytes else util.UNICODE]
-            self.re_no_dir = _wcparse.RE_WIN_NO_DIR[util.BYTES if self.is_bytes else util.UNICODE]
-        else:
-            self.sep = b'/' if self.is_bytes else '/'
-            self.seps = (self.sep,)
-            self.re_pathlib_norm = _RE_PATHLIB_DOT_NORM[util.BYTES if self.is_bytes else util.UNICODE]
-            self.re_no_dir = _wcparse.RE_NO_DIR[util.BYTES if self.is_bytes else util.UNICODE]
-        self._parse_patterns(pattern)
+        self.raw_chars = bool(self.flags & RAWCHARS)  # type: bool
+        self.follow_links = bool(self.flags & FOLLOW)  # type: bool
+        self.dot = bool(self.flags & DOTMATCH)  # type: bool
+        self.unix = not bool(self.flags & FORCEWIN)  # type: bool
+        self.negate = bool(self.flags & NEGATE)  # type: bool
+        self.globstar = bool(self.flags & GLOBSTAR)  # type: bool
+        self.braces = bool(self.flags & BRACE)  # type: bool
+        self.matchbase = bool(self.flags & MATCHBASE)  # type: bool
+        self.case_sensitive = _wcparse.get_case(self.flags)  # type: bool
+        self.limit = limit  # type: int
 
-        if (
-            (self.is_bytes and not isinstance(self.root_dir, bytes)) or
-            (not self.is_bytes and not isinstance(self.root_dir, str))
-        ):
+        forcewin = self.flags & FORCEWIN
+        if isinstance(pats[0], bytes):
+            ptype = util.BYTES
+            self.current = b'.'  # type: AnyStr
+            self.specials = (b'.', b'..')  # type: Tuple[AnyStr, ...]
+            self.empty = b''  # type: AnyStr
+            self.stars = b'**'  # type: AnyStr
+            self.sep = b'\\' if forcewin else b'/'  # type: AnyStr
+            self.seps = (b'/', self.sep) if forcewin else (self.sep,)  # type: Tuple[AnyStr, ...]
+            self.re_pathlib_norm = cast(Pattern[AnyStr], _RE_WIN_PATHLIB_DOT_NORM[ptype])  # type: Pattern[AnyStr]
+            self.re_no_dir = cast(Pattern[AnyStr], _wcparse.RE_WIN_NO_DIR[ptype])  # type: Pattern[AnyStr]
+        else:
+            ptype = util.UNICODE
+            self.current = '.'
+            self.specials = ('.', '..')
+            self.empty = ''
+            self.stars = '**'
+            self.sep = '\\' if forcewin else '/'
+            self.seps = ('/', self.sep) if forcewin else (self.sep,)
+            self.re_pathlib_norm = cast(Pattern[AnyStr], _RE_WIN_PATHLIB_DOT_NORM[ptype])
+            self.re_no_dir = cast(Pattern[AnyStr], _wcparse.RE_WIN_NO_DIR[ptype])
+
+        temp = os.fspath(root_dir) if root_dir is not None else self.current
+        if not isinstance(temp, type(pats[0])):
             raise TypeError(
                 'Pattern and root_dir should be of the same type, not {} and {}'.format(
-                    type(pattern[0]), type(self.root_dir)
+                    type(pats[0]), type(temp)
                 )
             )
 
-    def _iter_patterns(self, patterns):
+        self.root_dir = temp  # type: AnyStr
+        self._parse_patterns(pats)
+
+    def _iter_patterns(self, patterns: Sequence[AnyStr]) -> Iterator[Tuple[bool, AnyStr]]:
         """Iterate expanded patterns."""
 
         seen = set()
@@ -482,22 +501,20 @@ class Glob(object):
                     current_limit -= count
                     if current_limit < 1:
                         current_limit = 1
-        except _wcparse.bracex.ExpansionLimitException:
+        except bracex.ExpansionLimitException:
             raise _wcparse.PatternLimitException(
                 "Pattern limit exceeded the limit of {:d}".format(self.limit)
             )
 
-    def _parse_patterns(self, patterns):
+    def _parse_patterns(self, patterns: Sequence[AnyStr]) -> None:
         """Parse patterns."""
 
-        self.pattern = []
-        self.npatterns = []
         for is_neg, p in self._iter_patterns(patterns):
             if is_neg:
                 # Treat the inverse pattern as a normal pattern if it matches, we will exclude.
                 # This is faster as compiled patterns usually compare the include patterns first,
                 # and then the exclude, but glob will already know it wants to include the file.
-                self.npatterns.append(_wcparse._compile(p, self.negate_flags))
+                self.npatterns.append(cast(Pattern[AnyStr], _wcparse._compile(p, self.negate_flags)))
             else:
                 self.pattern.append(_GlobSplit(p, self.flags).split())
 
@@ -519,22 +536,22 @@ class Glob(object):
         ):
             self.nounique = True
 
-    def _is_hidden(self, name):
+    def _is_hidden(self, name: AnyStr) -> bool:
         """Check if is file hidden."""
 
         return not self.dot and name[0:1] == self.specials[0]
 
-    def _is_this(self, name):
+    def _is_this(self, name: AnyStr) -> bool:
         """Check if "this" directory `.`."""
 
         return name == self.specials[0] or name == self.sep
 
-    def _is_parent(self, name):
+    def _is_parent(self, name: AnyStr) -> bool:
         """Check if `..`."""
 
         return name == self.specials[1]
 
-    def _match_excluded(self, filename, is_dir):
+    def _match_excluded(self, filename: AnyStr, is_dir: bool) -> bool:
         """Check if file should be excluded."""
 
         if is_dir and not filename.endswith(self.sep):
@@ -548,21 +565,21 @@ class Glob(object):
 
         return matched
 
-    def _is_excluded(self, path, is_dir):
+    def _is_excluded(self, path: AnyStr, is_dir: bool) -> bool:
         """Check if file is excluded."""
 
-        return self.npatterns and self._match_excluded(path, is_dir)
+        return bool(self.npatterns and self._match_excluded(path, is_dir))
 
-    def _match_literal(self, a, b=None):
+    def _match_literal(self, a: AnyStr, b: Optional[AnyStr] = None) -> bool:
         """Match two names."""
 
         return a.lower() == b if not self.case_sensitive else a == b
 
-    def _get_matcher(self, target):
+    def _get_matcher(self, target: Optional[Union[AnyStr, Pattern[AnyStr]]]) -> Optional[Callable[..., Any]]:
         """Get deep match."""
 
         if target is None:
-            matcher = None
+            matcher = None  # type: Optional[Callable[..., Any]]
         elif isinstance(target, (str, bytes)):
             # Plain text match
             if not self.case_sensitive:
@@ -575,7 +592,7 @@ class Glob(object):
             matcher = target.match
         return matcher
 
-    def _lexists(self, path):
+    def _lexists(self, path: AnyStr) -> bool:
         """Check if file exists."""
 
         if not self.dir_fd:
@@ -587,7 +604,7 @@ class Glob(object):
         else:
             return True
 
-    def prepend_base(self, path):
+    def prepend_base(self, path: AnyStr) -> AnyStr:
         """Join path to base if pattern is not absolute."""
 
         if self.is_abs_pattern:
@@ -595,13 +612,13 @@ class Glob(object):
         else:
             return os.path.join(self.root_dir, path)
 
-    def _iter(self, curdir, dir_only, deep):
+    def _iter(self, curdir: Optional[AnyStr], dir_only: bool, deep: bool) -> Iterator[Tuple[AnyStr, bool, bool, bool]]:
         """Iterate the directory."""
 
         try:
-            fd = None
+            fd = None  # type: Optional[int]
             if self.is_abs_pattern and curdir:
-                scandir = curdir
+                scandir = curdir  # type: Union[AnyStr, int]
             elif self.dir_fd is not None:
                 fd = scandir = os.open(
                     os.path.join(self.root_dir, curdir) if curdir else self.root_dir,
@@ -616,10 +633,10 @@ class Glob(object):
                 yield special, True, True, False
 
             try:
-                with os.scandir(scandir) as scan:
+                with os.scandir(scandir) as scan:  # type: ignore[type-var]
                     for f in scan:
                         try:
-                            hidden = self._is_hidden(f.name)
+                            hidden = self._is_hidden(f.name)  # type: ignore[arg-type]
                             is_dir = f.is_dir()
                             if is_dir:
                                 is_link = f.is_symlink()
@@ -627,7 +644,7 @@ class Glob(object):
                                 # We don't care if a file is a link
                                 is_link = False
                             if (not dir_only or is_dir):
-                                yield f.name, is_dir, hidden, is_link
+                                yield f.name, is_dir, hidden, is_link  # type: ignore[misc]
                         except OSError:  # pragma: no cover
                             pass
             finally:
@@ -637,7 +654,13 @@ class Glob(object):
         except OSError:  # pragma: no cover
             pass
 
-    def _glob_dir(self, curdir, matcher, dir_only=False, deep=False):
+    def _glob_dir(
+        self,
+        curdir: AnyStr,
+        matcher: Optional[Callable[..., Any]],
+        dir_only: bool = False,
+        deep: bool = False
+    ) -> Iterator[Tuple[AnyStr, bool]]:
         """Recursive directory glob."""
 
         files = list(self._iter(curdir, dir_only, deep))
@@ -655,7 +678,7 @@ class Glob(object):
             if deep and not hidden and is_dir and follow:
                 yield from self._glob_dir(path, matcher, dir_only, deep)
 
-    def _glob(self, curdir, this, rest):
+    def _glob(self, curdir: AnyStr, part: _GlobPart, rest: List[_GlobPart]) -> Iterator[Tuple[AnyStr, bool]]:
         """
         Handle glob flow.
 
@@ -669,10 +692,10 @@ class Glob(object):
         - `globstar` `**`.
         """
 
-        is_magic = this.is_magic
-        dir_only = this.dir_only
-        target = this.pattern
-        is_globstar = this.is_globstar
+        is_magic = part.is_magic
+        dir_only = part.dir_only
+        target = part.pattern
+        is_globstar = part.is_globstar
 
         if is_magic and is_globstar:
             # Glob star directory `**`.
@@ -730,7 +753,7 @@ class Glob(object):
                 else:
                     yield path, is_dir
 
-    def _get_starting_paths(self, curdir, dir_only):
+    def _get_starting_paths(self, curdir: AnyStr, dir_only: bool) -> List[Tuple[AnyStr, bool]]:
         """
         Get the starting location.
 
@@ -751,7 +774,7 @@ class Glob(object):
             results = [(curdir, True)]
         return results
 
-    def is_unique(self, path):
+    def is_unique(self, path: AnyStr) -> bool:
         """Test if path is unique."""
 
         if self.nounique:
@@ -763,20 +786,20 @@ class Glob(object):
             unique = True
         return unique
 
-    def _pathlib_norm(self, path):
+    def _pathlib_norm(self, path: AnyStr) -> AnyStr:
         """Normalize path as `pathlib` does."""
 
         path = self.re_pathlib_norm.sub(self.empty, path)
         return path[:-1] if len(path) > 1 and path[-1:] in self.seps else path
 
-    def format_path(self, path, is_dir, dir_only):
+    def format_path(self, path: AnyStr, is_dir: bool, dir_only: bool) -> Iterator[AnyStr]:
         """Format path."""
 
         path = os.path.join(path, self.empty) if dir_only or (self.mark and is_dir) else path
         if self.is_unique(self._pathlib_norm(path) if self.pathlib else path):
             yield path
 
-    def glob(self):
+    def glob(self) -> Iterator[AnyStr]:
         """Starts off the glob iterator."""
 
         curdir = self.current
@@ -829,26 +852,56 @@ class Glob(object):
                             yield from self.format_path(match, is_dir, dir_only)
 
 
-def iglob(patterns, *, flags=0, root_dir=None, dir_fd=None, limit=_wcparse.PATTERN_LIMIT):
+def iglob(
+    patterns: Union[str, bytes, Sequence[AnyStr]],
+    *,
+    flags: int = 0,
+    root_dir: Optional[Union[AnyStr, 'os.PathLike[AnyStr]']] = None,
+    dir_fd: Optional[int] = None,
+    limit: int = _wcparse.PATTERN_LIMIT
+) -> Iterator[AnyStr]:
     """Glob."""
 
-    yield from Glob(util.to_tuple(patterns), flags, root_dir, dir_fd, limit).glob()
+    if not isinstance(patterns, (str, bytes)) and not patterns:
+        return
+
+    yield from Glob(patterns, flags, root_dir, dir_fd, limit).glob()
 
 
-def glob(patterns, *, flags=0, root_dir=None, dir_fd=None, limit=_wcparse.PATTERN_LIMIT):
+def glob(
+    patterns: Union[str, bytes, Sequence[AnyStr]],
+    *,
+    flags: int = 0,
+    root_dir: Optional[Union[AnyStr, 'os.PathLike[AnyStr]']] = None,
+    dir_fd: Optional[int] = None,
+    limit: int = _wcparse.PATTERN_LIMIT
+) -> List[AnyStr]:
     """Glob."""
 
     return list(iglob(patterns, flags=flags, root_dir=root_dir, dir_fd=dir_fd, limit=limit))
 
 
-def translate(patterns, *, flags=0, limit=_wcparse.PATTERN_LIMIT):
+def translate(
+    patterns: Union[str, bytes, Sequence[AnyStr]],
+    *,
+    flags: int = 0,
+    limit: int = _wcparse.PATTERN_LIMIT
+) -> Tuple[List[AnyStr], List[AnyStr]]:
     """Translate glob pattern."""
 
     flags = _flag_transform(flags)
-    return _wcparse.translate(patterns, flags, limit)
+    return _wcparse.translate(_wcparse.to_str_sequence(patterns), flags, limit)
 
 
-def globmatch(filename, patterns, *, flags=0, root_dir=None, dir_fd=None, limit=_wcparse.PATTERN_LIMIT):
+def globmatch(
+    filename: Union[AnyStr, 'os.PathLike[AnyStr]'],
+    patterns: Union[str, bytes, Sequence[AnyStr]],
+    *,
+    flags: int = 0,
+    root_dir: Optional[Union[AnyStr, 'os.PathLike[AnyStr]']] = None,
+    dir_fd: Optional[int] = None,
+    limit: int = _wcparse.PATTERN_LIMIT
+) -> bool:
     """
     Check if filename matches pattern.
 
@@ -856,45 +909,79 @@ def globmatch(filename, patterns, *, flags=0, root_dir=None, dir_fd=None, limit=
     but if `case_sensitive` is set, respect that instead.
     """
 
-    if root_dir is not None:
-        root_dir = os.fspath(root_dir)
+    pats = _wcparse.to_str_sequence(patterns)
 
+    # Shortcut out if we have no patterns
+    if not pats:
+        return False
+
+    ptype = type(pats[0])
+    rdir = os.fspath(root_dir) if root_dir is not None else root_dir
     flags = _flag_transform(flags)
-    filename = os.fspath(filename)
-    return _wcparse.compile(patterns, flags, limit).match(filename, root_dir, dir_fd)
+    fname = os.fspath(filename)
+
+    # Ensure types are not mismatched
+    if (rdir is not None and not isinstance(rdir, ptype)):
+        raise TypeError(
+            "Pattern type of '{}' does not match the resolved type of '{}' of the root dir".format(ptype, type(rdir))
+        )
+
+    return bool(_wcparse.compile(pats, flags, limit).match(fname, rdir, dir_fd))
 
 
-def globfilter(filenames, patterns, *, flags=0, root_dir=None, dir_fd=None, limit=_wcparse.PATTERN_LIMIT):
+def globfilter(
+    filenames: Iterable[Union[AnyStr, 'os.PathLike[AnyStr]']],
+    patterns: Union[str, bytes, Sequence[AnyStr]],
+    *,
+    flags: int = 0,
+    root_dir: Optional[Union[AnyStr, 'os.PathLike[AnyStr]']] = None,
+    dir_fd: Optional[int] = None,
+    limit: int = _wcparse.PATTERN_LIMIT
+) -> List[Union[AnyStr, 'os.PathLike[AnyStr]']]:
     """Filter names using pattern."""
 
-    if root_dir is not None:
-        root_dir = os.fspath(root_dir)
+    pats = _wcparse.to_str_sequence(patterns)
 
-    matches = []
+    # Shortcut out if we have no patterns
+    if not pats:
+        return []
+
+    ptype = type(pats[0])
+    rdir = os.fspath(root_dir) if root_dir is not None else root_dir
+
+    # Ensure types are not mismatched
+    if (rdir is not None and not isinstance(rdir, ptype)):
+        raise TypeError(
+            "Pattern type of '{}' does not match the resolved type of '{}' of the root dir".format(ptype, type(rdir))
+        )
+
+    matches = []  # type: List[Union[AnyStr, 'os.PathLike[AnyStr]']]
     flags = _flag_transform(flags)
-    obj = _wcparse.compile(patterns, flags, limit)
+    obj = _wcparse.compile(pats, flags, limit)
 
     for filename in filenames:
         temp = os.fspath(filename)
-        if obj.match(temp, root_dir, dir_fd):
+        if obj.match(temp, rdir, dir_fd):
             matches.append(filename)
     return matches
 
 
 @util.deprecated("This function will be removed in 9.0.")
-def raw_escape(pattern, unix=None, raw_chars=True):
+def raw_escape(pattern: AnyStr, unix: Optional[bool] = None, raw_chars: bool = True) -> AnyStr:
     """Apply raw character transform before applying escape."""
 
-    return _wcparse.escape(util.norm_pattern(pattern, False, raw_chars, True), unix=unix, pathname=True, raw=True)
+    return _wcparse.escape(
+        util.norm_pattern(pattern, False, raw_chars, True), unix=unix, pathname=True, raw=True
+    )
 
 
-def escape(pattern, unix=None):
+def escape(pattern: AnyStr, unix: Optional[bool] = None) -> AnyStr:
     """Escape."""
 
     return _wcparse.escape(pattern, unix=unix)
 
 
-def is_magic(pattern, *, flags=0):
+def is_magic(pattern: AnyStr, *, flags: int = 0) -> bool:
     """Check if the pattern is likely to be magic."""
 
     flags = _flag_transform(flags)
